﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;

namespace LockingToDeadlock {

	class Employee {
		public int Id { get; set; }

		public long Salary { get; set; }

		private Employee _manager;
		public Employee Manager {
			get => _manager;
			set {
				_manager = value;
				_manager.ManagedEmployees.Add(this);
			}	
		}

		public List<Employee> ManagedEmployees { get; } = new();

		public override string ToString() {
			var sb = new StringBuilder();
			
			sb.Append(ManagedEmployees.Count == 0 ? "Employee" : (Manager != null ? "Manager" : "Head-manager"));
			sb.Append($" {{ Id: {Id}, Salary: {Salary}, Manager.Id: {(Manager?.Id)?.ToString() ?? "self-managed"} ");
			if (ManagedEmployees.Count != 0) {
				sb.Append("ManagedEmployees = { ");
				foreach (var e in ManagedEmployees) {
					sb.Append(e.Id);
					sb.Append(' ');
				}
				sb.Append("} ");
			}
			sb.Append("}");

			return sb.ToString();
		}
	}

	class Program {
		const int ManagedEmployeeCount = 10;
		const int ManagerCount = 3;

		static List<Employee> Employees = new();
		static List<Employee> Managers = new();

		static Random GlobalRandom = new Random(42);

		static volatile int EmployeesAwarded = 0;
		static volatile int DepartmentsPenalised = 0;

		static void InitializeRandomEmployees() {
			var r = GlobalRandom;

			for (int i = 0; i < ManagerCount; i++) {
				Managers.Add(new Employee { Salary = r.Next(5000, 10000) });
			}

			for (int i = 0; i < ManagedEmployeeCount; i++) {
				var employee = new Employee { Salary = r.Next(1000, 2000), Manager = Managers[r.Next(Managers.Count)] };
				Employees.Add(employee);
			}

			for (int i = 0; i < Managers.Count; i++) {
				var manager = Managers[i];
				if (i < Managers.Count - 1) {
					manager.Manager = Managers[r.Next(i + 1, Managers.Count)];
				}
				Employees.Insert(r.Next(Employees.Count), manager);
			}

			for (int i = 0; i < Employees.Count; i++) {
				Employees[i].Id = i;
			}
		}

#if V1
		/// <summary>
		/// If an employee is selected as valued-employee, then his/hers salary is raised, and his/hers manager gets a raise too.
		/// </summary>
		static void AwardingThread() {
			Random r;

			lock (GlobalRandom) {
				r = new Random(GlobalRandom.Next());
			}

			while (true) {
				var valuedEmployee = Employees[r.Next(Employees.Count)];
				if (valuedEmployee.Manager == null) continue;

				lock (valuedEmployee) {
					valuedEmployee.Salary += 100;
				}

				lock (valuedEmployee.Manager) {
					valuedEmployee.Manager.Salary += 1000;
				}

				Interlocked.Increment(ref EmployeesAwarded);
			}
		}

		/// <summary>
		/// If there is a low performing department represented by its manager, then the manager and all his/hers employees will get their salaries lowered.
		/// </summary>
		static void PenaltyThread() {
			Random r;

			lock (GlobalRandom) {
				r = new Random(GlobalRandom.Next());
			}

			while (true) {
				var problemDepartmentManager = Managers[r.Next(Managers.Count)];

				lock (problemDepartmentManager) {
					problemDepartmentManager.Salary -= 1000;

					foreach (var departmentEmployee in problemDepartmentManager.ManagedEmployees) {
						lock (departmentEmployee) {
							departmentEmployee.Salary -= 100;
						}
					}
				}

				Interlocked.Increment(ref DepartmentsPenalised);
			}
		}
#endif

#if V1B
		/// <summary>
		/// If an employee is selected as valued-employee, then his/hers salary is raised, and his/hers manager gets a raise too.
		/// </summary>
		static void AwardingThread() {
			Random r;

			lock (GlobalRandom) {
				r = new Random(GlobalRandom.Next());
			}

			while (true) {
				var valuedEmployee = Employees[r.Next(Employees.Count)];
				if (valuedEmployee.Manager == null) continue;

				var manager = valuedEmployee.Manager;
				if (manager.Id < valuedEmployee.Id) {
					Monitor.Enter(manager);
					Monitor.Enter(valuedEmployee);
				} else {
					Monitor.Enter(valuedEmployee);
					Monitor.Enter(manager);
				}

				valuedEmployee.Salary += 100;
				valuedEmployee.Manager.Salary += 1000;

				Monitor.Exit(manager);
				Monitor.Exit(valuedEmployee);
				
				Interlocked.Increment(ref EmployeesAwarded);
			}
		}

		/// <summary>
		/// If there is a low performing department represented by its manager, then the manager and all his/hers employees will get their salaries lowered.
		/// </summary>
		static void PenaltyThread() {
			Random r;

			lock (GlobalRandom) {
				r = new Random(GlobalRandom.Next());
			}

			while (true) {
				var problemDepartmentManager = Managers[r.Next(Managers.Count)];

				// CORRECT LOCKING ORDER NEEDS TO BE ENFORCED HERE TOO!!!

				lock (problemDepartmentManager) {
					problemDepartmentManager.Salary -= 1000;

					foreach (var departmentEmployee in problemDepartmentManager.ManagedEmployees) {
						lock (departmentEmployee) {
							departmentEmployee.Salary -= 100;
						}
					}
				}

				Interlocked.Increment(ref DepartmentsPenalised);
			}
		}
#endif

#if V1C
		/// <summary>
		/// If an employee is selected as valued-employee, then his/hers salary is raised, and his/hers manager gets a raise too.
		/// </summary>
		static void AwardingThread() {
			Random r;

			lock (GlobalRandom) {
				r = new Random(GlobalRandom.Next());
			}

			while (true) {
				var valuedEmployee = Employees[r.Next(Employees.Count)];
				if (valuedEmployee.Manager == null) continue;

				var manager = valuedEmployee.Manager;
				if (manager.Id < valuedEmployee.Id) {
					Monitor.Enter(manager);
					Monitor.Enter(valuedEmployee);
				} else {
					Monitor.Enter(valuedEmployee);
					Monitor.Enter(manager);
				}

				valuedEmployee.Salary += 100;
				valuedEmployee.Manager.Salary += 1000;

				Monitor.Exit(manager);
				Monitor.Exit(valuedEmployee);
				
				Interlocked.Increment(ref EmployeesAwarded);
			}
		}

		/// <summary>
		/// If there is a low performing department represented by its manager, then the manager and all his/hers employees will get their salaries lowered.
		/// </summary>
		static void PenaltyThread() {
			Random r;

			lock (GlobalRandom) {
				r = new Random(GlobalRandom.Next());
			}

			var employeesToLock = new List<Employee>();
			while (true) {
				var problemDepartmentManager = Managers[r.Next(Managers.Count)];

				// Lock all employees:

				employeesToLock.Add(problemDepartmentManager);
				lock (problemDepartmentManager) {
					employeesToLock.AddRange(problemDepartmentManager.ManagedEmployees);
				}

				employeesToLock.Sort((e1, e2) => e1.Id.CompareTo(e2.Id));

				foreach (var e in employeesToLock) {
					Monitor.Enter(e);
				}

				// Penalise department atomically:

				foreach (var departmentEmployee in employeesToLock) {   // Note: We cannot enumerate problemDepartmentManager.ManagedEmployees here, as it might have changed since we acquired employee list for locking.
																		//		 If problemDepartmentManager.ManagedEmployees now contains other employees than we in have employeesToLock, by accessing employees from
																		//		 problemDepartmentManager.ManagedEmployees we would access Employee instances without holding their lock.
					if (departmentEmployee == problemDepartmentManager) {
						departmentEmployee.Salary -= 1000;
					} else {
						departmentEmployee.Salary -= 100;
					}
				}

				// Unlock all employees:

				foreach (var e in employeesToLock) {
					Monitor.Exit(e);
				}
				employeesToLock.Clear();    // Let's recycle the same List for other iteration to save GC allocations. We expect every department to have roughly the same amount of employees.

				// Update statistics:

				Interlocked.Increment(ref DepartmentsPenalised);
			}
		}
#endif

#if V1C_TYPO
		/// <summary>
		/// If an employee is selected as valued-employee, then his/hers salary is raised, and his/hers manager gets a raise too.
		/// </summary>
		static void AwardingThread() {
			Random r;

			lock (GlobalRandom) {
				r = new Random(GlobalRandom.Next());
			}

			while (true) {
				var valuedEmployee = Employees[r.Next(Employees.Count)];
				if (valuedEmployee.Manager == null) continue;

				var manager = valuedEmployee.Manager;
				if (manager.Id < valuedEmployee.Id) {
					Monitor.Enter(manager);
					Monitor.Enter(valuedEmployee);
				} else {
					Monitor.Enter(valuedEmployee);
					Monitor.Enter(manager);
				}

				valuedEmployee.Salary += 100;
				valuedEmployee.Manager.Salary += 1000;

				Monitor.Exit(manager);
				Monitor.Enter(valuedEmployee);	// TYPO HERE - Monitor.Enter instead of Monitor.Exit !!!
				
				Interlocked.Increment(ref EmployeesAwarded);
			}
		}

		/// <summary>
		/// If there is a low performing department represented by its manager, then the manager and all his/hers employees will get their salaries lowered.
		/// </summary>
		static void PenaltyThread() {
			Random r;

			lock (GlobalRandom) {
				r = new Random(GlobalRandom.Next());
			}

			var employeesToLock = new List<Employee>();
			while (true) {
				var problemDepartmentManager = Managers[r.Next(Managers.Count)];

				// Lock all employees:

				employeesToLock.Add(problemDepartmentManager);
				lock (problemDepartmentManager) {
					employeesToLock.AddRange(problemDepartmentManager.ManagedEmployees);
				}

				employeesToLock.Sort((e1, e2) => e1.Id.CompareTo(e2.Id));

				foreach (var e in employeesToLock) {
					Monitor.Enter(e);
				}

				// Penalise department atomically:

				foreach (var departmentEmployee in employeesToLock) {   // Note: We cannot enumerate problemDepartmentManager.ManagedEmployees here, as it might have changed since we acquired employee list for locking.
																		//		 If problemDepartmentManager.ManagedEmployees now contains other employees than we in have employeesToLock, by accessing employees from
																		//		 problemDepartmentManager.ManagedEmployees we would access Employee instances without holding their lock.
					if (departmentEmployee == problemDepartmentManager) {
						departmentEmployee.Salary -= 1000;
					} else {
						departmentEmployee.Salary -= 100;
					}
				}

				// Unlock all employees:

				foreach (var e in employeesToLock) {
					Monitor.Exit(e);
				}
				employeesToLock.Clear();    // Let's recycle the same List for other iteration to save GC allocations. We expect every department to have roughly the same amount of employees.

				// Update statistics:

				Interlocked.Increment(ref DepartmentsPenalised);
			}
		}
#endif

#if V2
		/// <summary>
		/// If an employee is selected as valued-employee, then his/hers salary is raised, and his/hers manager gets a raise too.
		/// </summary>
		static void AwardingThread() {
			Random r;

			lock (GlobalRandom) {
				r = new Random(GlobalRandom.Next());
			}

			while (true) {
				var valuedEmployee = Employees[r.Next(Employees.Count)];
				if (valuedEmployee.Manager == null) continue;

				lock (valuedEmployee) {
					valuedEmployee.Salary += 100;

					lock (valuedEmployee.Manager) {
						valuedEmployee.Manager.Salary += 1000;
					}
				}

				Interlocked.Increment(ref EmployeesAwarded);
			}
		}

		/// <summary>
		/// If there is a low performing department represented by its manager, then the manager and all his/hers employees will get their salaries lowered.
		/// </summary>
		static void PenaltyThread() {
			Random r;

			lock (GlobalRandom) {
				r = new Random(GlobalRandom.Next());
			}

			while (true) {
				var problemDepartmentManager = Managers[r.Next(Managers.Count)];

				Employee[] departmentEmployees;

				lock (problemDepartmentManager) {
					problemDepartmentManager.Salary -= 1000;

					departmentEmployees = problemDepartmentManager.ManagedEmployees.ToArray();  // ManagedEmployees List is mutable, so make a private copy.
				}

				foreach (var departmentEmployee in departmentEmployees) {
					lock (departmentEmployee) {
						departmentEmployee.Salary -= 100;
					}
				}

				Interlocked.Increment(ref DepartmentsPenalised);
			}
		}
#endif

		static void Main(string[] args) {
			InitializeRandomEmployees();

			Console.WriteLine("Initial state:");
			foreach (var e in Employees) {
				Console.WriteLine(e);
			}
			Console.WriteLine();

			Console.WriteLine($"*** Main thread sees: {EmployeesAwarded} EmployeesAwarded, {DepartmentsPenalised} DepartmentsPenalised, starting AwardingThread and entering Sleep() ...");

			var t1 = new Thread(AwardingThread) { IsBackground = true };
			t1.Start();

			Thread.Sleep(500);
			Console.WriteLine($"*** Main thread sees: {EmployeesAwarded} EmployeesAwarded, {DepartmentsPenalised} DepartmentsPenalised, entering Sleep() ...");
			Thread.Sleep(500);
			Console.WriteLine($"*** Main thread sees: {EmployeesAwarded} EmployeesAwarded, {DepartmentsPenalised} DepartmentsPenalised, starting PenaltyThread ...");

			var t2 = new Thread(PenaltyThread) { IsBackground = true };
			t2.Start();

			int i = 0;
			while (!Console.KeyAvailable) {
				Console.WriteLine($"*** {i} *** Main thread sees: {EmployeesAwarded} EmployeesAwarded, {DepartmentsPenalised} DepartmentsPenalised, entering Sleep() before another update ...");
				Thread.Sleep(500);
				i++;
			}

			_ = Console.ReadKey();

			Console.WriteLine();
			Console.WriteLine("Current state:");
			foreach (var e in Employees) {
				Console.WriteLine(e);
			}
		}
	}
}
